import codecs
import json
import os
import platform
import shlex
import subprocess
from configparser import ConfigParser
from importlib import util as importlib_util
from pathlib import Path

from yawsso import TRACE, Constant, logger


def xu(path):
    if str(path).startswith('~'):
        return os.path.expanduser(path)
    else:
        return path


def halt(error):
    logger.error(error)
    exit(1)


def invoke(cmd):
    try:
        output = subprocess.check_output(shlex.split(cmd), stderr=subprocess.STDOUT).decode()
        success = True
    except subprocess.CalledProcessError as e:
        output = e.output.decode()
        success = False
    return success, output.strip('\n')


class Poll(object):

    def __init__(self, cmd, output=True):
        self.cmd = cmd
        self.output = output

        self._proc = subprocess.Popen(
            shlex.split(self.cmd),
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
        )

    def _out(self, line):
        if line != "" and self.output:  # pragma: no cover
            logger.info(line)  # pragma: no

    def start(self):
        while True:
            line = self._proc.stdout.readline()
            if not line:
                break
            line = line.rstrip('\n')  # pragma: no cover
            self._out(line)  # pragma: no cover
        return self

    def resolve(self):
        success = True
        for line in self._proc.stderr.readlines():
            line = line.rstrip('\n')
            if line != "":
                logger.error(line)
                success = False
        return success


def list_directory(path):
    file_paths = []
    if os.path.exists(path):
        file_paths = Path(path).iterdir()
    file_paths = sorted(file_paths, key=os.path.getmtime)
    file_paths.reverse()  # sort by recently updated
    return [str(f) for f in file_paths]


def load_json(path):
    try:
        with open(path) as context:
            return json.load(context)
    except ValueError:
        logger.log(TRACE, f"Exception occur when loading JSON: {path}. Skip.")


def read_config(path):
    config = ConfigParser()
    config.read(path)
    return config


def write_config(path, config):
    with open(path, "w") as destination:
        config.write(destination)


def encrypt(obj: str):
    return codecs.encode(obj, encoding=str(Constant.ROT_13.value))


def decrypt(obj):
    return codecs.decode(obj, encoding=str(Constant.ROT_13.value))


class Exporter(object):
    """
    https://docs.aws.amazon.com/cli/latest/userguide/cli-configure-envvars.html
    """

    def __init__(self, credentials):
        self.credentials = credentials
        self.clipboard = ""

    def _make_powershell(self):
        self.clipboard = f"$Env:AWS_ACCESS_KEY_ID=\"{self.credentials['accessKeyId']}\"\n"
        self.clipboard += f"$Env:AWS_SECRET_ACCESS_KEY=\"{self.credentials['secretAccessKey']}\"\n"
        self.clipboard += f"$Env:AWS_SESSION_TOKEN=\"{self.credentials['sessionToken']}\""

    def _make_cmd(self):
        self.clipboard = f"set AWS_ACCESS_KEY_ID={self.credentials['accessKeyId']}\n"
        self.clipboard += f"set AWS_SECRET_ACCESS_KEY={self.credentials['secretAccessKey']}\n"
        self.clipboard += f"set AWS_SESSION_TOKEN={self.credentials['sessionToken']}"

    def _make_nix(self):
        self.clipboard = f"export AWS_ACCESS_KEY_ID={self.credentials['accessKeyId']}\n"
        self.clipboard += f"export AWS_SECRET_ACCESS_KEY={self.credentials['secretAccessKey']}\n"
        self.clipboard += f"export AWS_SESSION_TOKEN={self.credentials['sessionToken']}"

    def _make_windows(self):
        if "$P$G" in os.getenv("PROMPT", ""):
            # cmd.exe
            logger.debug(f"Detected Windows platform with cmd.exe")
            self._make_cmd()
        else:
            # powershell
            logger.debug(f"Detected Windows platform with PowerShell")
            self._make_powershell()

    def get_export_cmd(self):
        if platform.system() == "Windows":
            self._make_windows()
        else:
            # Unix
            logger.debug(f"Detected Nix platform")
            self._make_nix()

        return self.clipboard


def get_export_vars(profile_name, credentials):
    if credentials is None:
        logger.warning(f"No appropriate credentials found for profile '{profile_name}'. "
                       f"Skip exporting it. Use --trace flag to see possible error causes.")
        return

    pyperclip_spec = importlib_util.find_spec("pyperclip")
    pyperclip_found = pyperclip_spec is not None

    clipboard = Exporter(credentials).get_export_cmd()

    if pyperclip_found:
        import pyperclip  # pragma: no cover
        pyperclip.copy(clipboard)  # pragma: no cover
        logger.info(f"Credentials copied to your clipboard for profile '{profile_name}'")  # pragma: no cover
    else:
        logger.debug("Clipboard module pyperclip is not installed, showing encrypted credentials on terminal instead")
        print(encrypt(clipboard))



import sys                                                                         # pragma: no cover

if sys.version_info >= (3, 6):                                                     # pragma: no cover
    from yawsso.cli import main                                                    # pragma: no cover
    main()                                                                         # pragma: no cover
else:
    python_version = ".".join(map(str, sys.version_info[:3]))                      # pragma: no cover
    print("This tool requires Python >=3.6. Found {}".format(python_version))      # pragma: no cover
    sys.exit(1)                                                                    # pragma: no cover



import argparse
import logging
import os
import shutil
import sys
from pathlib import Path

from yawsso import PROGRAM, TRACE, Constant, logger, core, utils, cmd


def boot():
    handler = logging.StreamHandler(sys.stdout)
    formatter = logging.Formatter('%(message)s')  # print UNIX friendly format for PIPE use case
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    logger.setLevel(logging.INFO)

    args = parser().parse_args()

    if args.version:
        logger.info(Constant.VERSION_HELP.value)
        exit(0)  # just version print, don't even bother all the rest

    if args.trace:
        formatter = logging.Formatter('%(asctime)s %(name)-12s %(levelname)-8s %(message)s')
        handler.setFormatter(formatter)
        logger.setLevel(TRACE)
        logger.log(TRACE, "Logging level: TRACE")

    if args.debug:
        formatter = logging.Formatter('%(asctime)s %(name)-12s %(levelname)-8s %(message)s')
        handler.setFormatter(formatter)
        logger.setLevel(logging.DEBUG)
        logger.debug("Logging level: DEBUG")

    if args.bin:
        core.aws_bin = args.bin

    logger.log(TRACE, f"args: {args}")
    logger.log(TRACE, f"AWS_CONFIG_FILE: {core.aws_config_file}")
    logger.log(TRACE, f"AWS_SHARED_CREDENTIALS_FILE: {core.aws_shared_credentials_file}")
    logger.log(TRACE, f"AWS_SSO_CACHE_PATH: {core.aws_sso_cache_path}")
    logger.log(TRACE, f"Cache SSO JSON files: {utils.list_directory(core.aws_sso_cache_path)}")

    verify_files_exist()
    verify_aws_bin()

    return cmd.Command(args=args)


def parser():
    ap = argparse.ArgumentParser(
        prog=PROGRAM, description="Sync all named profiles when calling without any arguments"
    )
    ap.add_argument("--default", action="store_true", help="Sync AWS default profile and all named profiles")
    ap.add_argument("--default-only", action="store_true", help="Sync AWS default profile only and exit")
    ap.add_argument("-p", "--profiles", nargs="*", metavar="", help="Sync specified AWS named profiles")
    ap.add_argument("-b", "--bin", metavar="", help="AWS CLI v2 binary location (default to `aws` in PATH)")
    ap.add_argument("-d", "--debug", help="Debug output", action="store_true")
    ap.add_argument("-t", "--trace", help="Trace output", action="store_true")
    ap.add_argument("-e", "--export-vars", dest="export_vars1", help="Print out AWS ENV vars", action="store_true")
    ap.add_argument("-v", "--version", help="Print version and exit", action="store_true")

    sp = ap.add_subparsers(title="available commands", metavar="", dest="command")
    login_help = "Invoke aws sso login and sync all named profiles"
    login_description = f"{login_help}\nUse `default` profile or `AWS_PROFILE` if optional argument `--profile` absent"
    login_command = sp.add_parser(
        "login", description=login_description, help=login_help, formatter_class=argparse.RawDescriptionHelpFormatter
    )
    login_command.add_argument("-e", "--export-vars", help="Print out AWS ENV vars", action="store_true")
    login_command.add_argument("--profile", help="Login profile (use `default` or `AWS_PROFILE` if absent)", metavar="")
    login_command.add_argument("--this", action="store_true", help="Only sync this login profile")
    sp.add_parser("encrypt", help=f"Encrypt ({Constant.ROT_13.value.upper()}) stdin and exit")
    sp.add_parser("decrypt", help=f"Decrypt ({Constant.ROT_13.value.upper()}) stdin and exit")
    sp.add_parser("version", help="Print version and exit")

    return ap


def verify_files_exist():
    if not os.path.exists(core.aws_shared_credentials_file):
        logger.debug(f"{core.aws_shared_credentials_file} file does not exist. Attempting to create one.")
        try:
            Path(os.path.dirname(core.aws_shared_credentials_file)).mkdir(parents=True, exist_ok=True)
            with open(core.aws_shared_credentials_file, "w"):
                pass
        except Exception as e:
            logger.debug(f"Can not create {core.aws_shared_credentials_file}. Exception: {e}")
            utils.halt(f"{core.aws_shared_credentials_file} file does not exist. Please create one and try again.")

    if not os.path.exists(core.aws_config_file):
        utils.halt(f"{core.aws_config_file} does not exist")

    if not os.path.exists(core.aws_sso_cache_path):
        utils.halt(f"{core.aws_sso_cache_path} does not exist")


def verify_aws_bin():
    if shutil.which(core.aws_bin) is None:
        utils.halt(f"Can not find AWS CLI v2 `{core.aws_bin}` command.")

    cmd_aws_cli_version = f"{core.aws_bin} --version"
    aws_cli_success, aws_cli_version_output = utils.invoke(cmd_aws_cli_version)

    if not aws_cli_success:
        utils.halt(f"ERROR EXECUTING COMMAND: '{cmd_aws_cli_version}'. EXCEPTION: {aws_cli_version_output}")

    if "aws-cli/2" not in aws_cli_version_output:
        utils.halt(f"Required AWS CLI v2. Found {aws_cli_version_output}")

    logger.debug(aws_cli_version_output)


def export_tokens(co, profile_name, credentials):
    if co.export_vars:
        utils.get_export_vars(profile_name, credentials)


def sync_default(co):
    credentials = core.update_profile("default", co.config)
    export_tokens(co, "default", credentials)


def handle_flag_e(co):
    """Specific use case: making `yawsso -e` behaviour to sync default profile, print cred then exit"""
    if co.export_vars and not co.args.default and not co.args.profiles and not hasattr(co.args, 'profile'):
        sync_default(co)
        exit(0)


def handle_flag_default(co):
    """Specific use case: two flags to take care of default profile sync behaviour"""
    if co.args.default or co.args.default_only:
        sync_default(co)
        if co.args.default_only:
            exit(0)


def handle_update_profile_cases(co, n_profiles):
    profiles = []
    for np in co.args.profiles:
        if ":" in np:
            handle_profile_rename_on_sync(np, n_profiles, profiles, co)
        elif np.endswith("*"):
            handle_profile_prefix(np, n_profiles, profiles)
        else:
            handle_profile_normal_case(np, n_profiles, profiles)
    return profiles


def handle_flag_p(co):
    """Main use case: sync all named profiles"""
    n_profiles = list(map(lambda p: p.replace("profile ", ""), filter(lambda s: s != "default", co.config.sections())))
    logger.debug(f"Current named profiles in config: {str(n_profiles)}")

    core.profiles = n_profiles  # initialise with all named profiles first

    if co.args.profiles:
        profiles = handle_update_profile_cases(co, n_profiles)
        core.profiles = list(set(profiles))  # dedup
    # else sync all named profiles


def handle_profile_rename_on_sync(np, n_profiles, profiles, co):
    old, new = np.split(":")
    if old not in n_profiles:
        logger.warning(f"Named profile `{old}` is not specified in {core.aws_config_file}. Skipping...")
        return
    logger.debug(f"Renaming profile {old} to {new}")
    profiles.append(old)
    co.profiles_new_name[old] = new


def handle_profile_prefix(np, n_profiles, profiles):
    prefix = np.split("*")[0]
    logger.log(TRACE, f"Collecting all named profiles start with '{prefix}'")
    prefix_list = list(filter(lambda _p: _p.startswith(prefix), n_profiles))
    profiles.extend(prefix_list)


def handle_profile_normal_case(np, n_profiles, profiles):
    if np not in n_profiles:
        logger.warning(f"Named profile `{np}` is not specified in {core.aws_config_file}. Skipping...")
        return
    profiles.append(np)


def main():
    co = boot()

    if co.args.command:
        co.dispatch()  # subcommand dispatch

    # then continue with sync all named profiles below if subcommand does not happen to exit the program yet

    handle_flag_e(co)

    handle_flag_default(co)

    handle_flag_p(co)

    logger.debug(f"Syncing named profiles: {str(core.profiles)}")
    for profile_name in core.profiles:
        if profile_name in co.profiles_new_name:
            credentials = core.update_profile(profile_name, co.config, co.profiles_new_name[profile_name])
        else:
            credentials = core.update_profile(profile_name, co.config)

        export_tokens(co, profile_name, credentials)



"""yawsso cli package"""
import logging
from enum import Enum
from pathlib import Path

# versioning follow PEP440
__version__ = VERSION = '1.0.0rc2'  # pragma: no cover

PROGRAM = 'yawsso'  # pragma: no cover

TRACE = 5
logging.addLevelName(TRACE, 'TRACE')
logger = logging.getLogger(__name__)


class Constant(Enum):
    ROLE_CHAINING_DURATION_SECONDS = 3600
    AWS_SSO_CACHE_PATH = f"{Path.home()}/.aws/sso/cache"
    AWS_CONFIG_FILE = f"{Path.home()}/.aws/config"
    AWS_SHARED_CREDENTIALS_FILE = f"{Path.home()}/.aws/credentials"
    AWS_DEFAULT_REGION = "us-east-1"
    ROT_13 = "rot13"
    VERSION_HELP = f"{PROGRAM} {VERSION}"



import sys
from abc import ABC, abstractmethod

from yawsso import TRACE, Constant, logger, core, utils


class Command(object):

    def __init__(self, args):
        self.args = args
        self.config = utils.read_config(core.aws_config_file)
        self.profiles_new_name = dict()
        self.export_vars = self._build_export_vars()

    def _build_export_vars(self):
        """Make export_vars avail either side of subcommand"""
        x_vars = self.args.export_vars if hasattr(self.args, 'export_vars') and self.args.export_vars else False
        x_vars1 = self.args.export_vars1 if hasattr(self.args, 'export_vars1') and self.args.export_vars1 else False
        return x_vars or x_vars1

    def dispatch(self):
        if self.args.command == "version":
            logger.info(Constant.VERSION_HELP.value)
            exit(0)

        elif self.args.command == "encrypt":
            EncryptCommand(self).perform()

        elif self.args.command == "decrypt":
            DecryptCommand(self).perform()

        elif self.args.command == "login":
            LoginCommand(self).perform().handle()


class CommandAction(ABC):

    def __init__(self, co: Command):
        self.co = co

    @abstractmethod
    def perform(self):
        pass  # pragma: no cover


class EncryptCommand(CommandAction):

    def perform(self):
        for line in sys.stdin:
            print(utils.encrypt(line.rstrip("\n")))
        exit(0)


class DecryptCommand(CommandAction):

    def perform(self):
        for line in sys.stdin:
            print(utils.decrypt(line.rstrip("\n")))
        exit(0)


class LoginCommand(CommandAction):

    def __init__(self, co):
        super(LoginCommand, self).__init__(co)
        self.login_profile = "default"
        self.login_profile_new_name = ""

    def perform(self):
        cmd_aws_sso_login = f"{core.aws_bin} sso login"

        if self.co.args.profile:
            if ":" in self.co.args.profile:
                # support rename profile upon login then sync use case
                self.login_profile, self.login_profile_new_name = self.co.args.profile.split(":")
                self.co.profiles_new_name[self.login_profile] = self.login_profile_new_name
            else:
                self.login_profile = self.co.args.profile

            cmd_aws_sso_login = f"{cmd_aws_sso_login} --profile={self.login_profile}"

        logger.log(TRACE, f"Running command: `{cmd_aws_sso_login}`")

        login_success = utils.Poll(cmd_aws_sso_login, output=not self.co.export_vars).start().resolve()
        if not login_success:
            utils.halt(f"Error running command: `{cmd_aws_sso_login}`")

        return self

    def handle(self):
        """Handle is just centralised interface hook to exec all extended use cases or flags"""
        self._handle_flag_e()
        self._handle_flag_this()
        self._handle_flag_default()

    def _handle_flag_e(self):
        """
        Specific use case: making `yawsso login -e` or `yawsso login --profile NAME -e`
        to perform login, sync, print cred then exit
        """
        if self.co.export_vars:
            credentials = core.update_profile(self.login_profile, self.co.config, self.login_profile_new_name)
            utils.get_export_vars(self.login_profile, credentials)
            exit(0)

    def _handle_flag_this(self):
        if self.co.args.this:
            core.update_profile(self.login_profile, self.co.config, self.login_profile_new_name)
            exit(0)

    def _handle_flag_default(self):
        if self.login_profile == "default" and not self.co.export_vars:
            core.update_profile("default", self.co.config, self.login_profile_new_name)



from setuptools import setup

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="yawsso",
    # version="MOVE TO setup.cfg",
    description="Yet Another AWS SSO - sync up AWS CLI v2 SSO login session to legacy CLI v1 credentials",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/victorskl/yawsso",
    author="Victor San Kho Lin",
    author_email="victor@sankholin.com",
    license="MIT",
    packages=["yawsso"],
    zip_safe=False,
    entry_points={
        "console_scripts": ["yawsso=yawsso.cli:main"],
    },
    project_urls={
        "Bug Tracker": "https://github.com/victorskl/yawsso/issues",
    },
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    extras_require={
        "all": [
            "pyperclip",
        ],
        "test": [
            "pytest",
            "pytest-cov",
            "flake8",
            "mockito",
            "cli-test-helpers",
            "coveralls",
            "pyperclip",
        ],
        "dev": [
            "twine",
            "setuptools",
            "wheel",
            "build",
            "aws-cdk-lib",
            "constructs",
            "tox",
            "nose2",
            "pre-commit",
            "detect-secrets",
            "ggshield",
        ],
    },
    python_requires=">=3.6",
)



import copy
import json
import os
from configparser import NoSectionError
from datetime import datetime, timezone

from yawsso import TRACE, Constant, logger, utils as u

aws_bin = "aws"  # assume `aws` command avail in PATH and is v2. otherwise, allow mutation with -b flag
profiles = None
aws_sso_cache_path = u.xu(os.getenv("AWS_SSO_CACHE_PATH", Constant.AWS_SSO_CACHE_PATH.value))
aws_config_file = u.xu(os.getenv("AWS_CONFIG_FILE", Constant.AWS_CONFIG_FILE.value))
aws_shared_credentials_file = u.xu(os.getenv("AWS_SHARED_CREDENTIALS_FILE", Constant.AWS_SHARED_CREDENTIALS_FILE.value))
aws_default_region = os.getenv("AWS_DEFAULT_REGION", Constant.AWS_DEFAULT_REGION.value)


def get_aws_cli_v2_sso_cached_login(profile):
    file_paths = u.list_directory(aws_sso_cache_path)
    for file_path in file_paths:
        if not file_path.endswith('.json'):
            logger.log(TRACE, f"Not JSON file, skip: {file_path}")
            continue

        data = u.load_json(file_path)
        if data.get("startUrl") != profile["sso_start_url"]:
            logger.log(TRACE, f"Not equal SSO start url, skip: {file_path}")
            continue
        logger.log(TRACE, f"Using cached SSO login: {file_path}")
        return data


def update_aws_cli_v1_credentials(profile_name, profile, credentials):
    if credentials is None:
        logger.warning(f"No appropriate credentials found for profile '{profile_name}'. "
                       f"Skip syncing it. Use --trace flag to see possible error causes.")
        return

    region = profile.get("region", aws_default_region)
    config = u.read_config(aws_shared_credentials_file)
    if config.has_section(profile_name):
        config.remove_section(profile_name)
    config.add_section(profile_name)
    config.set(profile_name, "region", region)
    config.set(profile_name, "aws_access_key_id", credentials["accessKeyId"])
    config.set(profile_name, "aws_secret_access_key", credentials["secretAccessKey"])
    config.set(profile_name, "aws_session_token", credentials["sessionToken"])
    config.set(profile_name, "aws_security_token", credentials["sessionToken"])
    ts_expires_millisecond = credentials["expiration"]
    dt_utc = str(datetime.utcfromtimestamp(ts_expires_millisecond / 1000.0).isoformat() + '+0000')
    config.set(profile_name, "aws_session_expiration", dt_utc)
    u.write_config(aws_shared_credentials_file, config)

    logger.debug(f"Done syncing AWS CLI v1 credentials using AWS CLI v2 SSO login session for profile `{profile_name}`")


def parse_sso_cached_login_expiry(cached_login):
    # older versions of aws-cli might use non-standard format with `UTC` instead of `Z`
    expires_at = cached_login["expiresAt"].replace('UTC', 'Z')
    datetime_format_in_sso_cached_login = "%Y-%m-%dT%H:%M:%SZ"
    expires_utc = datetime.strptime(expires_at, datetime_format_in_sso_cached_login)
    return expires_utc


def parse_assume_role_credentials_expiry(dt_str):
    datetime_format_in_assume_role_expiration = "%Y-%m-%dT%H:%M:%S+00:00"
    expires_utc = datetime.strptime(dt_str, datetime_format_in_assume_role_expiration)
    return expires_utc


def parse_credentials_file_session_expiry(dt_str):
    datetime_format_in_cred_file_aws_session_expiration = "%Y-%m-%dT%H:%M:%S+0000"  # 2020-06-14T17:13:26+0000
    expires_utc = datetime.strptime(dt_str, datetime_format_in_cred_file_aws_session_expiration)
    return expires_utc


def parse_role_name_from_role_arn(role_arn):
    arr = role_arn.split('/')
    return str(arr[len(arr) - 1]).replace('"', '').replace("'", "")


def append_cli_global_options(cmd: str, profile: dict):
    ca_bundle = profile.get('ca_bundle', None)
    if ca_bundle:
        cmd = f"{cmd} --ca-bundle '{ca_bundle}'"

    logger.log(TRACE, f"COMMAND: {cmd}")
    return cmd


def check_sso_cached_login_expires(profile_name, profile):
    cached_login = get_aws_cli_v2_sso_cached_login(profile)
    if cached_login is None:
        u.halt(f"Can not find valid AWS CLI v2 SSO login cache in {aws_sso_cache_path} for profile {profile_name}.")

    expires_utc = parse_sso_cached_login_expiry(cached_login)

    if datetime.utcnow() > expires_utc:
        u.halt(f"Current cached SSO login is expired since {expires_utc.astimezone().isoformat()}. Try login again.")

    return cached_login


def fetch_credentials(profile_name, profile):
    cached_login = check_sso_cached_login_expires(profile_name, profile)

    cmd_get_role_cred = f"{aws_bin} sso get-role-credentials " \
                        f"--output json " \
                        f"--profile {profile_name} " \
                        f"--region {profile['sso_region']} " \
                        f"--role-name {profile['sso_role_name']} " \
                        f"--account-id {profile['sso_account_id']} " \
                        f"--access-token {cached_login['accessToken']}"

    cmd_get_role_cred = append_cli_global_options(cmd_get_role_cred, profile)

    role_cred_success, role_cred_output = u.invoke(cmd_get_role_cred)

    if not role_cred_success:
        logger.log(TRACE, f"ERROR EXECUTING COMMAND: '{cmd_get_role_cred}'. EXCEPTION: '{role_cred_output}'")
        return

    return json.loads(role_cred_output)['roleCredentials']


def get_role_max_session_duration(profile_name, profile):
    role_name = parse_role_name_from_role_arn(profile['role_arn'])

    cmd_get_role = f"{aws_bin} iam get-role " \
                   f"--output json " \
                   f"--profile {profile_name} " \
                   f"--role-name {role_name} " \
                   f"--region {profile['region']}"

    cmd_get_role = append_cli_global_options(cmd_get_role, profile)

    get_role_success, get_role_output = u.invoke(cmd_get_role)

    if not get_role_success:
        logger.log(TRACE, f"ERROR EXECUTING COMMAND: '{cmd_get_role}'. EXCEPTION: {get_role_output}")
        logger.debug(f"Can not determine role {role_name} maximum session duration. "
                     f"Using default value {Constant.ROLE_CHAINING_DURATION_SECONDS.value} seconds.")
        return Constant.ROLE_CHAINING_DURATION_SECONDS.value

    return json.loads(get_role_output)['Role']['MaxSessionDuration']


def fetch_credentials_with_assume_role(profile_name, profile):
    duration_seconds = get_role_max_session_duration(profile_name, profile)
    if duration_seconds > Constant.ROLE_CHAINING_DURATION_SECONDS.value:
        logger.log(TRACE, f"Role {profile['role_arn']} is configured with max duration `{duration_seconds}` seconds. "
                          f"But AWS SSO service-linked role to assume another role_arn defined in source_profile form "
                          f"`role chaining` (i.e. using a role to assume a second role). Fall back session duration "
                          f"to a maximum of one hour. Well, you can always `yawsso` again when session expired!")
        duration_seconds = Constant.ROLE_CHAINING_DURATION_SECONDS.value

    utc_now_ts = int(datetime.utcnow().replace(tzinfo=timezone.utc).timestamp())
    cmd_assume_role_cred = f"{aws_bin} sts assume-role " \
                           f"--output json " \
                           f"--profile {profile_name} " \
                           f"--role-arn {profile['role_arn']} " \
                           f"--role-session-name yawsso-session-{utc_now_ts} " \
                           f"--duration-seconds {duration_seconds} " \
                           f"--region {profile['region']}"

    cmd_assume_role_cred = append_cli_global_options(cmd_assume_role_cred, profile)

    role_cred_success, role_cred_output = u.invoke(cmd_assume_role_cred)

    if not role_cred_success:
        logger.log(TRACE, f"ERROR EXECUTING COMMAND: `{cmd_assume_role_cred}`. EXCEPTION: {role_cred_output}")
        return

    assume_role_cred = json.loads(role_cred_output)['Credentials']

    cred = {}
    cred.update(accessKeyId=assume_role_cred['AccessKeyId'])
    cred.update(secretAccessKey=assume_role_cred['SecretAccessKey'])
    cred.update(sessionToken=assume_role_cred['SessionToken'])

    expire_utc = parse_assume_role_credentials_expiry(assume_role_cred['Expiration'])
    expire_utc_ts_millisecond = int(expire_utc.replace(tzinfo=timezone.utc).timestamp() * 1000)
    cred.update(expiration=expire_utc_ts_millisecond)

    return copy.deepcopy(cred)


def eager_sync_source_profile(source_profile_name, source_profile):
    if profiles and source_profile_name in profiles:  # it will come in main loop, so no proactive sync required
        return
    config = u.read_config(aws_shared_credentials_file)
    if config.has_section(source_profile_name):
        cred_profile = dict(config.items(source_profile_name))
        session_expires_utc = parse_credentials_file_session_expiry(cred_profile['aws_session_expiration'])
        if datetime.utcnow() > session_expires_utc:
            logger.log(TRACE, f"Eagerly sync source_profile `{source_profile_name}`")
            credentials = fetch_credentials(source_profile_name, source_profile)
            update_aws_cli_v1_credentials(source_profile_name, source_profile, credentials)


def load_profile_from_config(profile_name, config):
    try:
        if profile_name == "default":
            profile_opts = config.items(f"{profile_name}")
        else:
            profile_opts = config.items(f"profile {profile_name}")
        return dict(profile_opts)
    except NoSectionError as e:
        u.halt(e)


def is_sso_profile(profile):
    return {"sso_start_url", "sso_account_id", "sso_role_name", "sso_region"} <= profile.keys()


def is_source_profile(profile):
    return {"source_profile", "role_arn", "region"} <= profile.keys()


def update_profile(profile_name, config, new_profile_name=""):
    profile = load_profile_from_config(profile_name, config)

    if new_profile_name == "":
        new_profile_name = profile_name
        logger.log(TRACE, f"Syncing profile... {profile_name}: {profile}")
    else:
        logger.log(TRACE, f"Syncing profile... {profile_name}->{new_profile_name}: {profile}")

    if is_sso_profile(profile):
        credentials = fetch_credentials(profile_name, profile)

    elif is_source_profile(profile):
        source_profile_name = profile['source_profile']
        source_profile = load_profile_from_config(source_profile_name, config)
        if not is_sso_profile(source_profile):
            logger.warning(f"Your source_profile is not an AWS SSO profile. Skip syncing profile `{profile_name}`")
            return
        check_sso_cached_login_expires(source_profile_name, source_profile)
        eager_sync_source_profile(source_profile_name, source_profile)
        logger.log(TRACE, f"Fetching credentials using assume role for `{profile_name}`")
        credentials = fetch_credentials_with_assume_role(profile_name, profile)

    else:
        logger.warning(f"Not an AWS SSO profile nor no source_profile found. Skip syncing profile `{profile_name}`")
        return

    update_aws_cli_v1_credentials(new_profile_name, profile, credentials)
    return credentials
